#include "matrix.h"
namespace std {

// ******************* CONSTRUCTORS *******************

// constructor
matrix::matrix(unsigned int rowNo, unsigned int colNo, double v) {
    if (!rowNo || !colNo)
        error("ERROR in matrix::matrix() : Zero matrix dimension!");
        
    this->rowNo = rowNo;
    this->colNo = colNo;

    vals = new vector<vector<double> >;
    vals->resize(rowNo);
	for(unsigned int r=0; r<=rowNo-1; ++r)
		(*vals)[r].resize(colNo,v);
}

// copy constructor
matrix::matrix(const matrix& m) {
    rowNo = m.rowNo;
    colNo = m.colNo;
    vals = new vector<vector<double> >;
    *vals = *m.vals;    // operator= of the STL vector class is used here
}

// ******************* DESTRUCTOR *******************

matrix::~matrix(void) {
    delete vals;
}

// ******************* FEW USEFUL FUNCTIONS *******************

// transpose function (modify & return *this object)
matrix & matrix::transpose() {
    matrix tmp(colNo,rowNo);

    for(unsigned int r=0; r<=rowNo-1; ++r)
        for(unsigned int c=0; c<=colNo-1; ++c)
	        tmp(c,r) = (*vals)[r][c];
    (*this) = tmp;  // operator= of our matrix class is used here

    return *this;
}

// randomise the matrix (modify & return *this object)
matrix & matrix::randomize(unsigned int limit) {
    for(unsigned int r=0; r<=rowNo-1; ++r)
        for(unsigned int c=0; c<=colNo-1; ++c)
	        (*vals)[r][c] = rand() % (limit+1);

    return *this;
}

// ******************* SELF-MODIFYING OPERATORS *******************

// assignment operator
matrix& matrix::operator= (const matrix& m) {
    rowNo = m.rowNo;
    colNo = m.colNo;
    *vals = *m.vals;    // operator= of the STL vector class is used here

    return *this;
}

// addition assignment operator
matrix& matrix::operator+= (const matrix& m) {
    if (rowNo!=m.rowNo || colNo!=m.colNo)
        error("ERROR in matrix::operator+= : Inconsistent matrix sizes!");

    for(unsigned int r=0; r<=rowNo-1; ++r)
        for(unsigned int c=0; c<=colNo-1; ++c)
	       (*vals)[r][c] += m(r,c);

    return *this;
}

// matrix multiplication assignment operator
matrix& matrix::operator*= (const matrix& m) {
    if (colNo != m.rowNo)
        error("ERROR in matrix::operator*= : Inconsistent matrix sizes!");

    matrix tmp(rowNo,m.colNo);

    for (unsigned int r=0; r<=rowNo-1; ++r)
        for (unsigned int c=0; c<=m.colNo-1; ++c) {
            tmp(r,c) = 0;
            for(unsigned int i=0; i<=colNo-1; ++i)
                tmp(r,c) += (*vals)[r][i]*m(i,c);
        }
    *this = tmp;

    return *this;
}

// scalar multiplication assignment operator
matrix& matrix::operator*= (const double& v) {
    for(unsigned int r=0; r<=rowNo-1; ++r)
	    for(unsigned int c=0; c<=colNo-1; ++c)
	       (*vals)[r][c] *= v;

    return *this;
}

// ******************* SUBSCRIPT OPERATORS *******************

// subscript operator to change individual elements 
double& matrix::operator() (unsigned int r, unsigned int c) {
    if (r>=rowNo || c>=colNo)
        error("ERROR in matrix::operator(): Index out of range!");

    return (*vals)[r][c];
}

// subscript operator to read individual elements (access only)
double matrix::operator()(unsigned int r, unsigned int c) const {
    if (r>=rowNo || c>=colNo)
        error("ERROR in matrix::operator(): Index out of range!");

    return (*vals)[r][c];
}

// ******************* INTERNAL (PRIVATE) FUNCTIONS *******************

void matrix::error(const char* pMsg) const {
    cerr <<pMsg <<endl <<endl;
    system("pause");    // for Dev C++ users only
    exit(1);
}

// **************************************************************************
// ******************* OVERLOADED NON-MODYFYING OPERATORS *******************
// **************************************************************************

// binary addition operator
matrix operator+ (const matrix & m1, const matrix & m2) {
    matrix tmp = m1;
    tmp += m2;
    return tmp;
}

// scalar binary multiplication operator (matrix * scalar)
matrix operator* (const matrix & m, const double & v) {
    matrix tmp = m;
    tmp *= v;
    return tmp;
}

// scalar binary multiplication operator (scalar * matrix)
matrix operator* (const double & v, const matrix & m) {
    return (m*v);
}

// matrix binary multiplication operator
matrix operator* (const matrix & m1, const matrix & m2) {
    matrix tmp = m1;
    tmp *= m2;
    return tmp;
}

// output stream function
ostream & operator << (ostream & ostr, const matrix & m)
{
    for(unsigned int r=0; r<=m.getRowNo()-1; ++r) {
        for(unsigned int c=0; c<=m.getColNo()-1; ++c) {
            ostr << m(r,c) << '\t';
        }
        ostr << endl;
    }
    return ostr;
}

} // namespace std
